from dandi.dandiapi import DandiAPIClient


def get_all_dandisets_metadata():
    with DandiAPIClient() as client:
        all_metadata = list()
        for ii, m in enumerate(client.get_dandisets()):
            if ii > 1 and ii < 560:
                try:
                    all_metadata.append(m.get_metadata())
                except:
                    pass
            else:
                pass
    return all_metadata


def get_dandiset_metadata(dandiset_id: str):
    with DandiAPIClient() as client:
        dandiset = client.get_dandiset(dandiset_id=dandiset_id, version_id="draft")
        return dandiset.get_metadata()


def list_dandiset_files(dandiset_id: str):
    with DandiAPIClient() as client:
        dandiset = client.get_dandiset(dandiset_id=dandiset_id, version_id="draft")
        return [i.dict().get("path") for i in dandiset.get_assets()]


def get_file_url(dandiset_id:str, file_path: str):
    with DandiAPIClient() as client:
        asset = client.get_dandiset(dandiset_id, 'draft').get_asset_by_path(file_path)
        return asset.get_content_url(follow_redirects=1, strip_query=True)


from datetime import datetime
from dateutil.tz import tzlocal

import ipywidgets as widgets

from pynwb import NWBFile

from nwbwidgets import nwb2widget
from nwbwidgets.panel import Panel


def test_nwbfile():
    nwbfile = NWBFile(
        "my first synthetic recording",
        "EXAMPLE_ID",
        datetime.now(tzlocal()),
        experimenter="Dr. Matthew Douglass",
        lab="Vision Neuroscience Laboratory",
        institution="University of Middle Earth at the Shire",
        experiment_description="We recorded from two macaque monkeys during memory-guided saccade task",
        session_id="LONELYMTL",
    )
    assert isinstance(nwb2widget(nwbfile), widgets.Widget)


def test_panel():
    panel = Panel()
    assert isinstance(panel, widgets.Widget)

    # Change dropdown options for coverage
    panel.source_options_radio.value = "Local dir"
    panel.source_options_radio.value = "Local file"
    panel.source_options_radio.value = "S3"
    panel.source_options_radio.value = "DANDI"

    # Choose DANDI set
    panel.source_dandi_id.value = panel.source_dandi_id.options[10]
    # Click accept specific file button
    panel.source_dandi_file_button.click()

    assert len(panel.widgets_panel.children) > 0


# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = 'NWB Widgets'
copyright = '2022, NeurodataWithoutBorders'
author = 'NeurodataWithoutBorders'

# -- Support building doc without install --------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
# sys.path.insert(0, os.path.abspath('.'))

import sys
import os

cwd = os.getcwd()
project_root = os.path.dirname(os.path.dirname(cwd))

# Insert the project root dir as the first element in the PYTHONPATH.
# This lets us ensure that the source package is imported, and that its
# version is used.
sys.path.insert(0, os.path.join(project_root, '../../nwbwidgets'))



# -- Autodoc configuration -----------------------------------------------------

autoclass_content = 'both'
autodoc_docstring_signature = True
autodoc_member_order = 'bysource'

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "myst_parser",  # Markdown support
    "sphinx.ext.napoleon",  # Support for NumPy and Google style docstrings
    "sphinx.ext.autodoc",  # Includes documentation from docstrings in docs/api
    "sphinx.ext.autosummary",  # To-add
    "sphinx_toggleprompt",  # Used to control >>> behavior in the conversion gallery example doctests
    "sphinx_copybutton",  # Used to control the copy button behavior in the conversion gallery doctsts
    "sphinx.ext.intersphinx",  # Allows links to other sphinx project documentation sites
    "sphinx_search.extension",  # Allows for auto search function the documentation
    "sphinx.ext.viewcode",  # Shows source code in the documentation
    "sphinx.ext.extlinks",  # Allows to use shorter external links defined in the extlinks variable.
]

templates_path = ['_templates']
exclude_patterns = []



# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = 'sphinx_rtd_theme'
html_static_path = ['_static']


# TODO - not working- Include arguments with typehints from classes
# ref: https://stackoverflow.com/questions/5599254/how-to-use-sphinxs-autodoc-to-document-a-classs-init-self-method
# ref: https://stackoverflow.com/questions/63957326/sphinx-does-not-show-class-instantiation-arguments-for-generic-classes-i-e-par
autoclass_content = 'both'
autodoc_typehints = 'signature'


import ipywidgets as widgets
from pathlib import Path
import h5py
import fsspec
from fsspec.implementations.cached import CachingFileSystem

from pynwb import NWBHDF5IO
from nwbwidgets import nwb2widget
from nwbwidgets.utils.dandi import (
    get_all_dandisets_metadata, 
    get_dandiset_metadata, 
    list_dandiset_files,
    get_file_url
)


class Panel(widgets.VBox):

    def __init__(
        self, 
        stream_mode: str = "fsspec", 
        cache_path: str = None, 
        enable_dandi_source: bool = True,
        enable_s3_source: bool = True,
        enable_local_source: bool = True,
        **kwargs
    ):
        """
        NWB widgets Panel for visualization of NWB files.

        Args:
            stream_mode (str, optional): Either "fsspec" or "ros3". Defaults to "fsspec".
            cache_path (str, optional): The path to cached data if streaming with "fsspec". If left as None, a directory "nwb-cache" is created under the current working directory. Defaults to None.
            enable_dandi_source (bool, optional): Enable DANDI source option. Defaults to True.
            enable_s3_source (bool, optional): Enable S3 source option. Defaults to True.
            enable_local_source (bool, optional): Enable local source option. Defaults to True.
        """
        super().__init__(children=[], **kwargs)

        self.stream_mode = stream_mode

        self.cache_path = cache_path
        if cache_path is None:
            self.cache_path = "nwb-cache"

        # Create a virtual filesystem based on the http protocol and use caching to save accessed data to RAM.
        if enable_dandi_source or enable_s3_source:
            self.cfs = CachingFileSystem(
                fs=fsspec.filesystem("http"),
                cache_storage=self.cache_path,  # Local folder for the cache
            )

        self.source_options_names = list()
        if enable_local_source:
            self.source_options_names.append("Local dir")
            self.source_options_names.append("Local file")
        if enable_dandi_source:
            self.source_options_names.append("DANDI")
        if enable_s3_source:
            self.source_options_names.append("S3")

        self.all_dandisets_metadata = None

        self.source_options_radio = widgets.RadioButtons(
            options=self.source_options_names, 
            layout=widgets.Layout(width='100px', overflow=None)
        )
        self.source_options_label = widgets.Label(
            'Source:',
            layout=widgets.Layout(width='100px', overflow=None)
        )
        self.source_options = widgets.VBox(
            children=[self.source_options_radio], 
            layout=widgets.Layout(width='120px', overflow=None, padding="5px 5px 5px 10px")
        )

        self.source_changing_panel = widgets.Box()

        self.input_form = widgets.HBox(
            children=[
                self.source_options, 
                self.source_changing_panel,
                # self.source_file_dandi_vbox
            ], 
            layout={"border": "1px solid gray"}
        )
        self.widgets_panel = widgets.VBox([])
        
        self.children = [self.input_form, self.widgets_panel]
        self.source_options_radio.observe(self.updated_source, 'value')

        if enable_local_source:
            self.source_options_radio.value = "Local dir"
            self.create_components_local_dir_source()
        elif enable_dandi_source:
            self.source_options_radio.value = "DANDI"
            self.create_components_dandi_source()


    def updated_source(self, args=None):
        """Update Panel components depending on chosen source."""
        if args['new'] == "DANDI":
            self.create_components_dandi_source()
        elif args['new'] == "S3":
            self.create_components_s3_source()
        elif args['new'] == "Local dir":
            self.create_components_local_dir_source()
        elif args['new'] == "Local file":
            self.create_components_local_file_source()
        
    
    def create_components_dandi_source(self, args=None):
        """Create widgets components for DANDI option"""
        if self.all_dandisets_metadata is None:
            self.source_changing_panel.children = [widgets.Label("Fetching DANDI datasets info, this might take up to a minute...")]
            self.all_dandisets_metadata = get_all_dandisets_metadata()
        
        dandiset_options = list()
        for m in self.all_dandisets_metadata:
            item_name = m.id.split(":")[1].split("/")[0] + " - " + m.name
            dandiset_options.append(item_name)

        self.source_dandi_id = widgets.Dropdown(options=dandiset_options, description='Dandiset:', layout=widgets.Layout(width='400px', overflow=None))
        self.source_dandi_file_dropdown = widgets.Dropdown(options=[], description='File:', layout=widgets.Layout(width='400px', overflow=None))
        self.source_dandi_file_button = widgets.Button(icon='check', description="Load file")

        self.source_dandi_vbox = widgets.VBox(
            children=[
                self.source_dandi_id, 
                self.source_dandi_file_dropdown,
                self.source_dandi_file_button
            ], 
            layout=widgets.Layout(padding="5px 0px 5px 0px")
        )

        self.dandi_summary = widgets.HTML(
            value="<style>p{word-wrap: break-word}</style> <p>" + "" + "</p>",
            layout=widgets.Layout(height="100px", width="700px", padding="5px 5px 5px 10px")
        )

        self.dandi_panel = widgets.HBox([self.source_dandi_vbox, self.dandi_summary])

        self.source_changing_panel.children = [self.dandi_panel]

        self.source_dandi_id.observe(self.list_dandiset_files_dropdown, 'value')
        self.source_dandi_file_button.on_click(self.stream_dandiset_file)
        self.list_dandiset_files_dropdown()


    def create_components_s3_source(self):
        """Create widgets components for S3 option"""
        self.source_s3_file_url = widgets.Text(
            value="",
            description="URL:",
        )
        self.source_s3_button = widgets.Button(icon='check', description="Load file")
        self.s3_panel = widgets.VBox(
            children=[
                self.source_s3_file_url, 
                self.source_s3_button
            ], 
            layout=widgets.Layout(padding="5px 0px 5px 0px")
        )
        self.source_changing_panel.children = [self.s3_panel]
        self.source_s3_button.on_click(self.stream_s3_file)


    def create_components_local_dir_source(self):
        """Create widgets components for Loca dir option"""
        self.local_dir_path = widgets.Text(
            value="",
            description="Dir path:",
            layout=widgets.Layout(width='400px', overflow=None)
        )
        self.local_dir_button = widgets.Button(description="Search")
        self.local_dir_top = widgets.HBox(
            children=[
                self.local_dir_path, 
                self.local_dir_button
            ], 
            layout=widgets.Layout(padding="5px 0px 5px 0px")
        )
        self.local_dir_files = widgets.Dropdown(options=[], description='Files:', layout=widgets.Layout(width='400px', overflow=None))
        self.local_dir_file_button = widgets.Button(icon='check', description="Load file")
        self.local_dir_panel = widgets.VBox(
            children=[
                self.local_dir_top,
                self.local_dir_files,
                self.local_dir_file_button
            ], 
            layout=widgets.Layout(padding="5px 0px 5px 0px")
        )
        self.source_changing_panel.children = [self.local_dir_panel]
        self.local_dir_button.on_click(self.list_local_dir_files)
        self.local_dir_file_button.on_click(self.load_local_dir_file)


    def create_components_local_file_source(self):
        """Create widgets components for Local file option"""
        self.local_file_path = widgets.Text(
            value="",
            description="File path:",
            layout=widgets.Layout(width='400px', overflow=None)
        )
        self.local_file_button = widgets.Button(icon='check', description="Load file")
        self.local_file_panel = widgets.VBox(
            children=[
                self.local_file_path,
                self.local_file_button
            ], 
            layout=widgets.Layout(padding="5px 0px 5px 0px")
        )
        self.source_changing_panel.children = [self.local_file_panel]
        self.local_file_button.on_click(self.load_local_file)


    def list_dandiset_files_dropdown(self, args=None):
        """Populate dropdown with all files and text area with summary"""
        self.dandi_summary.value = "Loading dandiset info..."
        self.source_dandi_file_dropdown.options = []      
        dandiset_id = self.source_dandi_id.value.split("-")[0].strip()
        self.source_dandi_file_dropdown.options = list_dandiset_files(dandiset_id=dandiset_id)

        metadata = get_dandiset_metadata(dandiset_id=dandiset_id)
        self.dandi_summary.value = "<style>p{word-wrap: break-word}</style> <p>" + metadata.description + "</p>"

    
    def list_local_dir_files(self, args=None):
        """List NWB files in local dir"""
        if Path(self.local_dir_path.value).is_dir():
            all_files = [f.name for f in Path(self.local_dir_path.value).glob("*.nwb")]
            self.local_dir_files.options = all_files
        else:
            print("Invalid local dir path")


    def stream_dandiset_file(self, args=None):
        """Stream NWB file from DANDI"""
        self.widgets_panel.children = [widgets.Label("loading...")]
        dandiset_id = self.source_dandi_id.value.split("-")[0].strip()
        file_path = self.source_dandi_file_dropdown.value
        s3_url = get_file_url(dandiset_id=dandiset_id, file_path=file_path)
        if self.stream_mode == "ros3":
            io = NWBHDF5IO(s3_url, mode='r', load_namespaces=True, driver='ros3')

        elif self.stream_mode == "fsspec":            
            f = self.cfs.open(s3_url, "rb")
            file = h5py.File(f)
            io = NWBHDF5IO(file=file, load_namespaces=True)

        nwbfile = io.read()
        self.widgets_panel.children = [nwb2widget(nwbfile)]
    

    def stream_s3_file(self, args=None):
        """Stream NWB file from S3 url"""
        self.widgets_panel.children = [widgets.Label("loading...")]
        s3_url = self.source_s3_file_url.value
        if self.stream_mode == "ros3":
            io = NWBHDF5IO(s3_url, mode='r', load_namespaces=True, driver='ros3')
        elif self.stream_mode == "fsspec":            
            f = self.cfs.open(s3_url, "rb")
            file = h5py.File(f)
            io = NWBHDF5IO(file=file, load_namespaces=True)

        nwbfile = io.read()
        self.widgets_panel.children = [nwb2widget(nwbfile)]


    def load_local_dir_file(self, args=None):
        """Load local NWB file"""
        full_file_path = str(Path(self.local_dir_path.value) / self.local_dir_files.value)
        io = NWBHDF5IO(full_file_path, mode='r', load_namespaces=True)
        nwb = io.read()
        self.widgets_panel.children = [nwb2widget(nwb)]


    def load_local_file(self, args=None):
        """Load local NWB file"""
        full_file_path = str(Path(self.local_file_path.value))
        io = NWBHDF5IO(full_file_path, mode='r', load_namespaces=True)
        nwb = io.read()
        self.widgets_panel.children = [nwb2widget(nwb)]
